use core::f32;
use std::time::Duration;

use bevy::{
    asset::RenderAssetUsages,
    pbr::MaterialExtension,
    picking::PickSet,
    prelude::*,
    render::mesh::{Indices, PrimitiveTopology},
};
use noise::{Fbm, MultiFractal, NoiseFn, Perlin};
use rand::{thread_rng, Rng, SeedableRng};
use voronoice::{BoundingBox, Point, VoronoiBuilder};

mod animals;
mod cells;
mod picking;
pub use animals::AnimalKind;
pub use cells::CellKind;
use cells::{
    material::{CellMaterial, CellMaterialExtension},
    *,
};
use picking::*;

pub struct Plugin;
impl bevy::prelude::Plugin for Plugin {
    fn build(&self, app: &mut App) {
        app.add_systems(Startup, setup)
            .add_systems(PreUpdate, picking_backend.in_set(PickSet::Backend))
            .add_systems(Update, (update_map_mesh, cells_regeneration, expand))
            .insert_resource(ClearColor(Color::srgb(0., 0., 1.)))
            .insert_resource(Seed(thread_rng().gen()))
            .add_plugins(MaterialPlugin::<CellMaterial>::default());
    }
}

pub const HEIGHT: f32 = 16.;
pub const WIDTH: f32 = 16.;
pub const REAL_HEIGHT: f32 = 8000.;
pub const REAL_WIDTH: f32 = 8000.;
pub const CELL_AREA: f32 = REAL_HEIGHT * REAL_WIDTH / SIZE as f32;
pub const SIZE: usize = 100_000;

#[derive(Resource)]
struct Seed(u32);

#[derive(Resource)]
pub struct Voronoi(voronoice::Voronoi);

#[derive(Component)]
pub struct MapMarker;

#[derive(Component)]
struct MapMeshColors(Vec<[f32; 4]>);

#[derive(Resource)]
pub struct CellsEntities(Vec<Entity>);

#[derive(Component)]
pub struct MeshNeedsUpdate(bool);

fn setup(
    mut cmds: Commands,
    mut meshes: ResMut<Assets<Mesh>>,
    // mut materials: ResMut<Assets<StandardMaterial>>,
    mut materials: ResMut<Assets<CellMaterial>>,
    seed: Res<Seed>,
) {
    let mut rng = rand::rngs::SmallRng::seed_from_u64(seed.0 as u64);
    let mut sites = Vec::with_capacity(SIZE);
    for _ in 0..SIZE {
        sites.push(Point {
            x: rng.gen_range(-WIDTH / 2.0..WIDTH / 2.0) as f64,
            y: rng.gen_range(-HEIGHT / 2.0..HEIGHT / 2.0) as f64,
        })
    }
    let voronoi = VoronoiBuilder::default()
        .set_sites(sites)
        .set_bounding_box(BoundingBox::new_centered(WIDTH as f64, HEIGHT as f64))
        .set_lloyd_relaxation_iterations(3)
        .build()
        .unwrap();
    let mut cells = Vec::with_capacity(SIZE);
    let z_noise = Fbm::<Perlin>::new(seed.0);
    let moisture_noise = Fbm::<Perlin>::new(seed.0 + 1).set_frequency(2.);
    for i in 0..SIZE {
        let c = voronoi.cell(i);
        let site = c.site_position();
        let z = get_altitude(&z_noise, &[site.x as f32, site.y as f32]);
        let _m = (
            (moisture_noise.get([site.x, site.y]) + 1.) / 2.
            // Noise + [0; 1]
        )
        .clamp(0., 1.) as f32;
        let k = if z <= 0. {
            CellKind::Sea
        } else if z <= 0.04 {
            CellKind::Beach
        } else if z < 0.6 {
            CellKind::Dirt
        } else {
            CellKind::Stone
        };
        cells.push(Cell {
            kind: k,
            voronoi_id: i,
            vertices: vec![],
        });
    }

    let mut poss = Vec::new();
    let mut indices = Vec::new();
    // let mut normals = Vec::new();

    for (i, pos) in voronoi.sites().iter().enumerate() {
        let z = get_altitude(&z_noise, &[pos.x as f32, pos.y as f32]);
        poss.push(Vec3::new(pos.x as f32, pos.y as f32, z));
        cells[i].vertices.push(i);
    }
    for t in voronoi.triangulation().triangles.chunks_exact(3) {
        let on_hull = t
            .iter()
            .filter(|v| voronoi.triangulation().hull.contains(v))
            .count();
        // Don't draw triangles on the hull that are often long and look glitchy
        if on_hull > 0 {
            continue;
        }
        indices.extend_from_slice(&[t[2] as u32, t[1] as u32, t[0] as u32]);
    }
    // indices.extend(voronoi.triangulation().triangles.iter().map(|t| *t as u32));

    // for (c, cd) in voronoi.iter_cells().zip(cells.iter_mut()) {
    //     let vs = c.iter_vertices().collect::<Vec<_>>();
    //     let i = poss.len();
    //     for v in vs.iter() {
    //         let z = get_altitude(&z_noise, &[v.x as f32, v.y as f32]);
    //         poss.push(Vec3::new(v.x as f32, v.y as f32, z));
    //         // const EPSILON: f32 = 0.01;
    //         // let dzx = get_altitude(&z_noise, &[v.x as f32 + EPSILON, v.y as f32]) - z;
    //         // let nx = (f32::consts::FRAC_PI_2 - (dzx / EPSILON).atan()).cos();
    //         // let dzy = get_altitude(&z_noise, &[v.x as f32, v.y as f32 + EPSILON]) - z;
    //         // let ny = (f32::consts::FRAC_PI_2 - (dzy / EPSILON).atan()).cos();

    //         // let nz = (1. - (nx.powi(2) + ny.powi(2))).sqrt();

    //         // normals.push(Vec3::new(nx, ny, nz));
    //     }
    //     for v in 1..(vs.len() - 1) {
    //         indices.extend_from_slice(&[(i + v + 1) as u32, (i + v) as u32, i as u32]);
    //         cd.vertices.extend_from_slice(&[i, i + v, i + v + 1]);
    //     }
    // }

    let colors = vec![[0.; 4]; poss.len()];
    let mut mesh = Mesh::new(
        PrimitiveTopology::TriangleList,
        RenderAssetUsages::default(),
    )
    // Add 4 vertices, each with its own position attribute (coordinate in
    // 3D space), for each of the corners of the parallelogram.
    .with_inserted_attribute(Mesh::ATTRIBUTE_POSITION, poss)
    .with_inserted_attribute(Mesh::ATTRIBUTE_COLOR, colors.clone())
    // .with_inserted_attribute(Mesh::ATTRIBUTE_NORMAL, normals)
    .with_inserted_indices(Indices::U32(indices));

    mesh.compute_smooth_normals();

    let mut cells_entities = Vec::with_capacity(cells.len());

    cmds.spawn((
        Mesh3d(meshes.add(mesh)),
        // StandardMaterial
        // MeshMaterial3d(materials.add(StandardMaterial::default())),
        MeshMaterial3d(materials.add(CellMaterial {
            base: StandardMaterial::default(),
            extension: CellMaterialExtension {},
        })),
        Transform::default(),
        MapMeshColors(colors),
        MeshNeedsUpdate(true),
        MapMarker,
    ))
    .with_children(|parent| {
        for cell in cells {
            let kind = cell.kind;
            let mut cmd = parent.spawn(cell);
            match kind {
                CellKind::Grass | CellKind::Forest => {
                    cmd.insert((
                        Wealth(0),
                        Regeneration {
                            last_update: Duration::ZERO,
                            full_growth_duration: kind.regen_full_growth_duration(),
                        },
                    ));
                }
                _ => {}
            }
            cmd.observe(self::cells::on_click);
            cells_entities.push(cmd.id());
        }
    });
    cmds.insert_resource(Voronoi(voronoi));
    cmds.insert_resource(CellsEntities(cells_entities));
}

fn update_map_mesh(
    cells: Query<(&Cell, Option<&Wealth>)>,
    mut map: Query<(&Mesh3d, &mut MapMeshColors, &mut MeshNeedsUpdate), With<MapMarker>>,
    mut meshes: ResMut<Assets<Mesh>>,
) {
    let (mesh, mut cols, mut needs_update) = map.single_mut();
    if needs_update.0 {
        if let Some(mesh) = meshes.get_mut(mesh) {
            let mut modified = false;
            // let mut rng = thread_rng();
            for (cell, wealth) in cells.iter() {
                // let col: [f32; 4] = [rng.gen(), rng.gen(), rng.gen(), 1.];
                let col = cell.color(wealth.map(|w| w.0).unwrap_or_default());
                for id in cell.vertices.iter() {
                    modified = modified || cols.0[*id] != col;
                    cols.0[*id] = col.clone();
                }
            }
            if modified {
                mesh.insert_attribute(Mesh::ATTRIBUTE_COLOR, cols.0.clone());
            }
        }
        needs_update.0 = false;
    }
}

/// Between 0 and 1
fn get_altitude(z_noise: &Fbm<Perlin>, pos: &[f32; 2]) -> f32 {
    let z_noise = ((z_noise.get([pos[0] as f64, pos[1] as f64]) as f32) + 1.) / 2.; // Noise [0; 1]
    (
        -0.2// Arbitrary value
        + (z_noise.exp() / 1f32.exp())
        // Noise + [0; 1]
        // - ((pos[0].powi(2)+pos[1].powi(2)).sqrt()*0.3)
        // Distance - [0; sqrt(2)] * 0.5
    )
    // .clamp(-1., 1.)
}
