use std::time::Duration;

use bevy::prelude::*;
use rand::{seq::IteratorRandom, thread_rng, Rng};

use crate::{time::GameTime, ui::CurrentAction};

use super::{animals::Animal, AnimalKind, CellsEntities, MapMarker, MeshNeedsUpdate, Voronoi};

pub mod material;

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum CellKind {
    Sea,
    Beach,
    Forest,
    Dirt,
    Stone,
    Grass,
}
impl CellKind {
    pub fn regen_full_growth_duration(&self) -> Duration {
        match self {
            CellKind::Sea | CellKind::Beach | CellKind::Dirt | CellKind::Stone => unreachable!(),
            CellKind::Forest => Duration::from_secs(100 * 365 * 24 * 60 * 60), // Let's say that a forest takes 100 years to mature
            CellKind::Grass => Duration::from_secs(7 * 7 * 24 * 60 * 60), // Let's say that grass takes 7 weeks to reach its max
        }
    }
    pub fn can_place_animal(&self, kind: AnimalKind) -> bool {
        match self {
            CellKind::Sea => false,
            _ => true,
        }
    }
}

#[derive(Debug, Component)]
pub struct Cell {
    pub kind: CellKind,
    pub voronoi_id: usize,
    pub vertices: Vec<usize>,
}
impl Cell {
    pub fn color(&self, wealth: u8) -> [f32; 4] {
        match self.kind {
            CellKind::Sea => [0., 0., 1., 1.],
            CellKind::Beach => [0.82, 0.84, 0.51, 1.],
            CellKind::Forest => [0., 0.5 - (wealth as f32 / 255. * 0.4), 0., 1.],
            CellKind::Dirt => [
                0.53 - (wealth as f32 / 255. * 0.4),
                0.38 - (wealth as f32 / 255. * 0.4),
                0.29 - (wealth as f32 / 255. * 0.4),
                1.,
            ],
            CellKind::Stone => [0.5, 0.5, 0.5, 1.],
            CellKind::Grass => [
                (136. / 255.) - (wealth as f32 / 255. * 0.15),
                (154. / 255.) + (wealth as f32 / 255. * 0.1),
                (59. / 255.) - (wealth as f32 / 255. * 0.15),
                1.,
            ],
        }
    }
}

#[derive(Component)]
pub struct Wealth(pub u8);

impl Default for Wealth {
    fn default() -> Self {
        Wealth(u8::MAX)
    }
}

#[derive(Component)]
#[require(Wealth)]
pub struct Regeneration {
    pub last_update: Duration,
    pub full_growth_duration: Duration,
}

pub fn cells_regeneration(
    mut cells: Query<(&mut Regeneration, &mut Wealth)>,
    mut map_needs_update: Query<&mut MeshNeedsUpdate, With<MapMarker>>,
    gt: Res<GameTime>,
) {
    let mut map_needs_update = map_needs_update.single_mut();
    for (mut regen, mut wealth) in cells.iter_mut() {
        if gt.current - regen.last_update > regen.full_growth_duration / u8::MAX as u32 {
            regen.last_update = gt.current;
            wealth.0 = wealth.0.saturating_add(1);
            map_needs_update.0 = true;
        }
    }
}

pub fn expand(
    mut cells: Query<(&mut Cell, Option<&Wealth>)>,
    voronoi: Res<Voronoi>,
    cells_entities: Res<CellsEntities>,
    mut cmds: Commands,
    t: Res<Time>,
    gt: Res<GameTime>,
) {
    let mut random = thread_rng();
    let mut changes = Vec::new();
    for (cell, wealth) in cells.iter() {
        // TODO: search beter numbers, for now they are arbitrary
        if cell.kind == CellKind::Forest || cell.kind == CellKind::Grass {
            let wealth = if cell.kind == CellKind::Forest {
                1.
            } else {
                wealth.unwrap().0 as f64 / 255.
            };
            if random.gen_bool(
                (t.elapsed_secs_f64() * (gt.speed as f64) * wealth / (60. * 60. * 24. * 30.))
                    .min(1.),
            ) {
                // Let say that grass takes 1 months to expand
                let target = voronoi
                    .0
                    .cell(cell.voronoi_id)
                    .iter_neighbors()
                    .choose(&mut random)
                    .unwrap();
                changes.push((target, CellKind::Grass));
            }
        }
        if cell.kind == CellKind::Forest {
            if random.gen_bool(
                (t.elapsed_secs_f64()
                    * (gt.speed as f64)
                    * (wealth.unwrap().0 as f64 / 255.).sqrt()
                    / (60. * 60. * 24. * 365. * 5.))
                    .min(1.),
            ) {
                // Let say that forest takes 5 years to expand
                let target = voronoi
                    .0
                    .cell(cell.voronoi_id)
                    .iter_neighbors()
                    .choose(&mut random)
                    .unwrap();
                changes.push((target, CellKind::Forest));
            }
        }
    }
    for (v_id, kind) in changes {
        let target = cells_entities.0[v_id];
        let (mut cell, _) = cells.get_mut(target).unwrap();
        if kind == CellKind::Forest && (cell.kind == CellKind::Dirt || cell.kind == CellKind::Grass)
        {
            cell.kind = CellKind::Forest;
            cmds.entity(target).insert((
                Wealth(0),
                Regeneration {
                    full_growth_duration: CellKind::Forest.regen_full_growth_duration(),
                    last_update: gt.current,
                },
            ));
        } else if kind == CellKind::Grass && cell.kind == CellKind::Dirt {
            cell.kind = CellKind::Grass;
            cmds.entity(target).insert((
                Wealth(0),
                Regeneration {
                    full_growth_duration: CellKind::Grass.regen_full_growth_duration(),
                    last_update: gt.current,
                },
            ));
        }
    }
}

pub fn on_click(
    trigger: Trigger<Pointer<Click>>,
    mut cells: Query<&mut Cell>,
    voronoi: Res<Voronoi>,
    mut map_needs_update: Query<&mut MeshNeedsUpdate, With<MapMarker>>,
    mut cmds: Commands,
    ca: Res<CurrentAction>,
    gt: Res<GameTime>,
) {
    if trigger.duration > Duration::from_millis(200) {
        return;
    }
    let mut cell = cells.get_mut(trigger.target).unwrap();
    match *ca {
        CurrentAction::ChangeCell(ck) => match ck {
            CellKind::Forest => match cell.kind {
                CellKind::Dirt | CellKind::Grass => {
                    cmds.entity(trigger.target).insert((
                        Wealth(0),
                        Regeneration {
                            last_update: gt.current,
                            full_growth_duration: CellKind::Forest.regen_full_growth_duration(),
                        },
                    ));
                    cell.kind = CellKind::Forest;
                    map_needs_update.single_mut().0 = true;
                }
                _ => {}
            },
            CellKind::Grass => match cell.kind {
                CellKind::Dirt => {
                    cmds.entity(trigger.target).insert((
                        Wealth(0),
                        Regeneration {
                            last_update: gt.current,
                            full_growth_duration: CellKind::Grass.regen_full_growth_duration(),
                        },
                    ));
                    cell.kind = CellKind::Grass;
                    map_needs_update.single_mut().0 = true;
                }
                _ => {}
            },
            _ => {}
        },
        CurrentAction::AddAnimal(ak) => {
            if cell.kind.can_place_animal(ak) {
                let v_cell = voronoi.0.cell(cell.voronoi_id);
                let cell_pos = v_cell.site_position();
                cmds.spawn((
                    Animal { kind: ak },
                    Transform::from_xyz(cell_pos.x as f32, cell_pos.y as f32, 0.),
                ));
            }
        }
        _ => {}
    }
}
